import logging
import socket
import threading
import time

from typing import Any
from typing import Callable
from typing import Optional
from typing import Union

from .protocol import Status
from .protocol import Connection


class RawSocket(Connection):
    def __init__(
        self,
        address: str,
        port: int,
        observer: Optional[Callable[[str, str], None]] = None  # FIXME: !
    ):
        self.log = logging.getLogger(__name__)
        self.log.addHandler(logging.NullHandler())

        self.observer_name = "CONNECTION"
        self.observer = observer if observer else lambda st, nd: None
        self.observer.post(self.observer_name, Status.DISCONNETCED)

        self.send_ready = threading.Lock()

        self.address = address
        self.port = port
        self.socket_timeout = 30_000
        self.RECEIVE_SIZE = 2048

        self.log.debug(f"Address: {self.address}")
        self.log.debug(f"Port: {self.port}")

        self._socket_setup()

    def _socket_setup(self) -> None:
        """
        Create socket to communicate with server.

        Creates a socket instance and sets the options for communication.
        Passes the socket to the ssl_wrap method

        Note:
        After 5 idle minutes, start sending keepalives every 1 minutes.
        Drop connection after 2 failed keepalives
        """
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.setsockopt(
            socket.SOL_SOCKET,
            socket.SO_KEEPALIVE,
            1
        )
        self.socket.settimeout(2)
        if (
            hasattr(socket, "TCP_KEEPIDLE")
            and hasattr(socket, "TCP_KEEPINTVL")
            and hasattr(socket, "TCP_KEEPCNT")
        ):
            self.log.debug(
                "Setting TCP_KEEPIDLE, TCP_KEEPINTVL & TCP_KEEPCNT."
            )
            self.socket.setsockopt(
                socket.SOL_TCP,
                socket.TCP_KEEPIDLE,
                5 * 60
            )
            self.socket.setsockopt(
                socket.IPPROTO_TCP,
                socket.TCP_KEEPIDLE,
                5 * 60
            )
            self.socket.setsockopt(
                socket.IPPROTO_TCP,
                socket.TCP_KEEPINTVL,
                60
            )
            self.socket.setsockopt(
                socket.IPPROTO_TCP,
                socket.TCP_KEEPCNT,
                2
            )

        if hasattr(socket, "TCP_USER_TIMEOUT"):
            self.log.debug(
                f"Setting TCP_USER_TIMEOUT to {self.socket_timeout}ms."
            )
            self.socket.setsockopt(
                socket.IPPROTO_TCP,
                socket.TCP_USER_TIMEOUT,
                self.socket_timeout
            )

    def send(
        self,
        data: Union[str, bytes]
    ) -> bool:
        """
        Send the str/Bytes to the server.

        If given string, it is encoded as 'uft-8' & send.

        UNSURE(MBK): Should the encoding be moved outside this class?

        Returns:
            True, if the data could be send else
            False.
        """

        if isinstance(data, str):
            data = data.encode('utf-8')

        try:
            self.socket.sendall(data)
        except ConnectionError:
            msg = "Get an ConnectionError, while trying to send"
            self.log.exception(msg)
            self.reconnect()
            return False
        except TimeoutError:
            msg = "Get an TimeoutError, while trying to send"
            self.log.exception(msg)
            # Reconnect?
            return False
        else:
            self.log.debug(f"Raw Data Send: {data}")
            return True

    def receive(self, parser: Callable[[bytes], Any]) -> Any:
        """
        Socket receive method.

        Method that handles receiving data from a socket. Capable of handling
        data chunks.

        Args:
            Callable: A parser, that returns the parsed data.
                      On Parsen Error, it should raise a
                      ValueError TypeError or any subClasses of those.
                      (Like 'JSONDecodeError' & 'pydantic.ValidationError' is)
        Returns:
            The "parser"'s output.

        """
        data = []
        while self.socket:
            try:
                data_chunk = self.socket.recv(self.RECEIVE_SIZE)
            except socket.timeout:
                continue

            if data_chunk == b'':
                self.log.debug("Server Closed socket.")
                self.reconnect()
            data.append(data_chunk)

            try:
                parsed_data = parser(b"".join(data))
            except ValueError:  # parentClass for JSONDecodeError.
                pass
            except TypeError:  # parentClass for pydantic.ValidationError
                pass
            else:
                self.log.debug(f"Raw Data Received: {data}")
                return parsed_data

    def connect(self) -> bool:
        """
        Connect to the server.

        Attempts a connection to the server on the provided addres and port.

        Returns:
            'True' if the connection was successful else
            'False'
        """

        try:
            self.log.info("Trying to Connect.")
            self.observer.post(self.observer_name, Status.CONNECTING)
            # self.socket.settimeout(10)  # Why?
            self.socket.connect((self.address, self.port))
            # self.socket.settimeout(None)  # Why?
            self.log.info(
                f"Connected on interface: {self.socket.getsockname()[0]}"
            )
            self.observer.post(self.observer_name, Status.CONNECTED)
            # if self.sockt_thread is None:
            #     self._start()
            return True

        except Exception as e:
            self.log.error("Failed to connect: {}".format(e))
            return False

    def reconnect(self, retry_limit: Optional[int] = None) -> bool:
        """
        Attempt to reconnect.

        Reconnect to the server, until the given amount af attempts,
        are above the retry_limit.
        if the retry_limit are not set, it will never end.

        Returns:
            'True' if the connection was successful else
            'False'
        """
        self.log.info("Reconnection...")

        while retry_limit is None or retry_limit > 0:
            if retry_limit:
                retry_limit -= 1
            self.disconnect()
            self._socket_setup()
            if self.connect():
                return True
            self.log.info("Trying to reconnect in 5 seconds")
            time.sleep(5)
        return False

    def disconnect(self) -> None:
        """Disconnect from the server."""
        if self.socket:
            self.socket.close()

    def close(self) -> None:
        """
        Close the connection.

        Closes the socket object connection.
        """
        self.log.info("Closing connection...")
        self.observer.post(self.observer_name, Status.DISCONNECTING)
        if self.socket:
            self.socket.close()
            self.socket = None
        self.observer.post(self.observer_name, Status.DISCONNETCED)
        self.log.info("Connection closed!")



from abc import ABC
from abc import abstractmethod

from typing import Any
from typing import Dict
from typing import Callable
# from typing import Union
# from typing import List
# from typing import TypeVar
from typing import Optional

from enum import Enum

from pydantic import BaseModel
from pydantic import UUID4
# from pydantic import HttpUrl

from ..service.template import ServiceClass

from ..schema.base_schema import PermissionType
from ..schema.base_schema import WappstoObject

# from ..schema.iot_schema import WappstoObjectType


# #############################################################################
#                             Value Settings Schema
# #############################################################################

class IoTEvent(str, Enum):
    CREATE = "create"  # POST
    CHANGE = "change"  # PUT
    REQUEST = "request"  # GET
    DELETE = "delete"  # DELETE


class ValueBaseType(str, Enum):
    """Internal use only!."""
    STRING = "string"
    NUMBER = "number"
    BLOB = "blob"
    XML = "xml"


class ValueType(str, Enum):
    """
    Predefined ValueTypes.

    Each of the predefined ValueTypes, have default
    value parameters set, which include BaseType, name,
    permission, range, step and the unit.
    """

    DEFAULT = "Default"
    STRING = "String"
    NUMBER = "Number"
    BLOB = "Blob"
    XML = "Xml"
    TEMPERATURE = "Temperature"  # TODO: !
    SPEED = "Speed"  # TODO: !
    BOOLEAN = "Boolean"
    LATITUDE = "Latitude"
    LONGITUDE = "Longitude"


class ValueSettinsSchema(BaseModel):
    type: ValueBaseType
    name: str
    permission: PermissionType
    mapping: Optional[Dict]  # Number only
    ordered_mapping: Optional[bool]  # Number only
    meaningful_zero: Optional[bool]  # Number only
    si_conversion: Optional[str]  # Number only
    min: Optional[int]  # Number only
    max: Optional[int]  # Blob, number, str only.
    step: Optional[int]  # Number only
    encoding: Optional[str]  # Blob, str only.
    xsd: Optional[str]  # XML only
    namespace: Optional[str]  # XML only
    unit: Optional[str]


valueSettings: Dict[ValueType, ValueSettinsSchema] = {
    ValueType.DEFAULT: ValueSettinsSchema(
        type=ValueBaseType.NUMBER,
        name="number",
        permission=PermissionType.READWRITE,
        mapping=None,  # dict,
        ordered_mapping=None,  # Boolean
        meaningful_zero=None,  # Boolean
        min=0,
        max=255,
        step=1,
        unit=None
    ),
    ValueType.STRING: ValueSettinsSchema(
        type=ValueBaseType.STRING,
        name="string",
        permission=PermissionType.READWRITE,
        max=64,
        encoding="utf-8",
        unit=None
    ),
    ValueType.NUMBER: ValueSettinsSchema(
        type=ValueBaseType.NUMBER,
        name="number",
        permission=PermissionType.READWRITE,
        mapping=None,  # dict,
        ordered_mapping=None,  # Boolean
        meaningful_zero=None,  # Boolean
        min=-1e+38,  # UNSURE(MBK): !!
        max=1e+38,
        step=1e-038,
        unit=None
    ),
    ValueType.BLOB: ValueSettinsSchema(
        type=ValueBaseType.BLOB,
        name="Blob",
        permission=PermissionType.READWRITE,
        max=64,
        encoding="base64",
        unit=None
    ),
    ValueType.XML: ValueSettinsSchema(
        type=ValueBaseType.XML,
        name="Xml",
        permission=PermissionType.READWRITE,
        xsd="",
        namespace="",
        unit=None
    ),
    ValueType.LATITUDE: ValueSettinsSchema(
        type=ValueBaseType.NUMBER,
        name="latitude",
        permission=PermissionType.READ,
        mapping=None,  # dict,
        ordered_mapping=None,  # Boolean
        meaningful_zero=True,  # Boolean
        min=-90,
        max=90,
        step=0.000001,
        unit="°N"
    ),
    ValueType.LONGITUDE: ValueSettinsSchema(
        type=ValueBaseType.NUMBER,
        name="longitude",
        permission=PermissionType.READ,
        mapping=None,  # dict,
        ordered_mapping=None,  # Boolean
        meaningful_zero=True,  # Boolean
        min=-180,
        max=180,
        step=0.000001,
        unit="°E"
    ),
    ValueType.BOOLEAN: ValueSettinsSchema(
        type=ValueBaseType.NUMBER,
        name="boolean",
        permission=PermissionType.READWRITE,
        mapping=None,  # dict,
        ordered_mapping=None,  # Boolean
        meaningful_zero=True,  # Boolean
        min=0,
        max=1,
        step=1,
        unit="Boolean"
    ),
}

# #############################################################################
#                             Config-File Schema
# #############################################################################


# class _UnitsInfo(BaseModel):
#     self_type: WappstoObjectType
#     parent: Optional[UUID4] = None
#     name: Optional[str] = None
#     self_id: int
#     children: List[UUID4]
#     children_id_mapping: Dict[int, UUID4]
#     children_name_mapping: Dict[str, UUID4]


# class _Config(BaseModel):
#     network_uuid: UUID4
#     # network_name: str  # Really needed?
#     port: int
#     end_point: HttpUrl
#     # connectSync: Optional[bool]
#     # storeQueue: Optional[bool]
#     # mixMaxEnforce: Optional[str]
#     # stepEnforce: Optional[str]
#     # deltaHandling: Optional[str]
#     # period_handling: Optional[str]


# class _ConfigFile(BaseModel):
#     configs: _Config
#     # NOTE: the str, should be UUID4, but can't do to pydantic error!
#     units: Dict[str, _UnitsInfo]


def dict_diff(olddict: Dict[Any, Any], newdict: Dict[Any, Any]):
    """Find & return what have updated from old to new dictionary."""
    return dict(set(newdict.items() - set(olddict.items())))


# #############################################################################
#                             Wappsto-Device Template
# #############################################################################

class WappstoUnit(ABC):

    schema: WappstoObject

    @property
    @abstractmethod
    def uuid(self) -> UUID4:
        pass

    # @abstractmethod
    # @property
    # def name(self) -> str:
    #     pass

    @property
    def id(self) -> int:
        pass

    @property
    @abstractmethod
    def children_uuid_mapping(self) -> Dict[UUID4, Callable]:
        pass

    @property
    @abstractmethod
    def children_id_mapping(self) -> Dict[int, UUID4]:
        pass

    @property
    @abstractmethod
    def children_name_mapping(self) -> Dict[str, UUID4]:
        pass

    @property
    @abstractmethod
    def connection(self) -> ServiceClass:
        pass

    # @abstractmethod
    # def _get_json(self) -> List[_UnitsInfo]:
    #     pass

    @abstractmethod
    def delete(self):
        pass

    @abstractmethod
    def change(self):
        pass

    @abstractmethod
    def onChange(self):
        pass

    @abstractmethod
    def onRequest(self):
        pass

    @abstractmethod
    def onRefresh(self):
        pass

    @abstractmethod
    def onDelete(self):
        pass



import uuid
import datetime

from enum import Enum
from itertools import zip_longest

from typing import Optional
from typing import Union

from pydantic import BaseModel
from pydantic import Extra
from pydantic import validator

from .base_schema import BlobValue
from .base_schema import Device
from .base_schema import Network
from .base_schema import NumberValue
from .base_schema import State
from .base_schema import StringValue
from .base_schema import XmlValue


def parwise(values):
    a = iter(values)
    return zip_longest(a, a)


ValueUnion = Union[StringValue, NumberValue, BlobValue, XmlValue]

JsonRpc_error_codes = {
    # Rpc Error Code: [HTTP Error Code, "Error String"]
    -32700: [400, "Parse error"],
    -32600: [400, "Invalid Request"],
    -32601: [404, "Method not found"],
    -32602: [400, "Invalid params"],
    -32603: [500, "Internal Server Error"],

    -32000: [404, "Timeout"],
    -32001: [301, "Moved Permanently"],
    -32002: [400, "Bad Request"],
    -32003: [401, "Unauthorized"],
    -32004: [402, "Payment Required"],
    -32005: [403, "Forbidden"],
    -32006: [404, "Not Found"],
    -32007: [405, "Method Not Allowed"],
    -32008: [406, "Not Acceptable"],
    -32009: [408, "Request Timeout"],
    -32010: [409, "Conflict"],
    -32011: [410, "Gone"],
    -32012: [415, "Unsupported Media Type"],
    -32013: [418, "I'm a teapot"],
    -32014: [501, "Not implemented"],
    -32015: [502, "Bad Gateway"],
    -32016: [503, "Service Unavailable"],
    -32017: [504, "Gateway Timeout"],
    -32018: [401, "Queue Limit Reached"]
}


class WappstoMethod(str, Enum):
    GET = "GET"
    POST = "POST"
    PATCH = "PATCH"
    PUT = "PUT"
    DELETE = "DELETE"
    HEAD = "HEAD"


class WappstoObjectType(str, Enum):
    # WappstoMetaType
    NETWORK = "network"
    DEVICE = "device"
    VALUE = "value"
    STATE = "state"


class Success(BaseModel):
    success: bool = True

    class Config:
        extra = Extra.forbid


class Identifier(BaseModel):
    identifier: str


class JsonMeta(BaseModel):
    server_send_time: datetime.datetime


class JsonReply(BaseModel):
    value: Optional[Union[
        Device,
        Network,
        State,
        ValueUnion,
        bool
    ]]
    meta: JsonMeta

    class Config:
        extra = Extra.forbid


class JsonData(BaseModel):
    data: Optional[Union[
        Device,
        Network,
        State,
        ValueUnion
    ]]
    url: str
    meta: Optional[Identifier]

    class Config:
        extra = Extra.forbid

    @validator('url')
    def path_check(cls, v, values, **kwargs):
        if v.startswith("/services/2.0/"):
            selftype, selfid = v.replace("/services/2.0/", "").split('/')
            WappstoObjectType(selftype)
            uuid.UUID(selfid)
        for selftype, selfid in parwise(v.split("/")[1:]):
            WappstoObjectType(selftype)
            if selfid:
                uuid.UUID(selfid)
        return v



import __main__
import netrc
import logging
import atexit
import json

from pathlib import Path

# from urllib.parse import ParseResult

from enum import Enum

from typing import Any, Dict, Optional, Union, Callable

from pydantic import UUID4
# from pydantic import parse_file_as


from ..service.template import ServiceClass
# from .service.rest_api import RestAPI
from ..service.iot_api import IoTAPI

from .template import dict_diff
from .device import Device
# from .value import Value

# from .template import _Config
# from .template import _ConfigFile
# from .template import _UnitsInfo

from ..schema import base_schema as WSchema
from ..schema.iot_schema import WappstoMethod

# from .utils import exceptions
from ..utils.certificateread import CertificateRead
from ..utils.offline_storage import OfflineStorage
from ..utils.offline_storage import OfflineStorageFiles


# #############################################################################
#                        Status Observer Setup
# #############################################################################

from ..utils import observer 
from ..connections import protocol as connection
from ..service import template as service


class StatusID(str, Enum):
    CONNECTION = "CONNECTION"
    SERVICE = "SERVICE"


# TODO: TEST the Status thingy.
ConnectionStatus = connection.Status.DISCONNETCED
ServiceStatus = service.Status.IDLE


def __connectionStatus(
    layer: StatusID,
    newStatus: connection.Status
):
    global ConnectionStatus
    ConnectionStatus = newStatus


def __serviceStatus(
    layer: StatusID,
    newStatus: service.Status
):
    global ServiceStatus
    ServiceStatus = newStatus


def subscribe_all_status():    
    observer.subscribe(
        event_name=StatusID.CONNECTION,
        callback=__connectionStatus
    )
    observer.subscribe(
        event_name=StatusID.SERVICE,
        callback=__serviceStatus
    )


# #############################################################################
#                                 Network Setup
# #############################################################################

class NetworkChangeType(str, Enum):
    description = "description"
    device = "device"
    info = "info"
    name = "name"


class NetworkRequestType(str, Enum):
    refresh = "refresh"
    delete = "delete"


class ConnectionTypes(str, Enum):
    IOTAPI = "jsonrpc"
    RESTAPI = "HTTPS"

# class status:
#     Connection -> *REf
#


class Network(object):

    schema = WSchema.Network

    def __init__(
        self,
        configFolder: Union[Path, str] = ".",  # Relative to the main.py-file.
        name: str = "TheNetwork",
        description: str = "",
        connection: ConnectionTypes = ConnectionTypes.IOTAPI,
        mixMaxEnforce="warning",  # "ignore", "enforce"
        stepEnforce="warning",  # "ignore", "enforce"
        deltaHandling="",
        period_handling="",
        connectSync: bool = True,  # Start with a Network GET to sync.  # TODO:
        pingPongPeriod: Optional[int] = None,  # Period between a RPC ping-pong.
        offlineStorage: Union[OfflineStorage, bool] = False,
        # none_blocking=True,  # Whether the post should wait for reply or not.
    ) -> None:
        """
        Configure the WappstoIoT settings.

        This function call is optional.
        If it is not called, the default settings will be used for WappstoIoT.
        This function will also connect to the WappstoIoT API on call.
        In the cases that this function is not called, the connection will be
        make when an action is make that requests the connection.

        The 'minMaxEnforce' is default set to "Warning" where a warning is
        reading to log, when the value range is outside the minimum & maximum
        range.
        The 'ignore' is where it do nothing when it is outside range.
        The 'enforce' is where the range are enforced to fit the minimum &
        maximum range. Meaning if it is above the maximum it is changed to
        the maximum, if it is below the minimum, it is set to the minimum value.
        """
        self.log = logging.getLogger(__name__)
        self.log.addHandler(logging.NullHandler())

        self.closed = False

        kwargs = locals()
        self.__uuid: UUID4
        self.element: WSchema.Network = self.schema()

        self.children_uuid_mapping: Dict[UUID4, Device] = {}
        self.children_id_mapping: Dict[int, UUID4] = {}
        self.children_name_mapping: Dict[str, UUID4] = {}

        self.cloud_id_mapping: Dict[int, UUID4] = {}

        if not isinstance(configFolder, Path):
            if configFolder == "." and hasattr(__main__, '__file__'):
                configFolder = Path(__main__.__file__).absolute().parent / Path(configFolder)
            else:
                configFolder = Path(configFolder)

        self.configFolder = configFolder

        self.connection: ServiceClass

        self._setup_offline_storage(offlineStorage)

        if connection == ConnectionTypes.IOTAPI:
            self._setup_IoTAPI(configFolder)
            cer = CertificateRead(crt=self.configFolder / "client.crt")
            self.__uuid = cer.network

        elif connection == ConnectionTypes.RESTAPI:
            # TODO: Find & load configs.
            configs: Dict[Any, Any] = {}
            self._setup_RestAPI(self.configFolder, configs)  # FIXME:

        subscribe_all_status()

        self.element = self.schema(
            name=name,
            description=description,
            meta=WSchema.NetworkMeta(
                version=WSchema.WappstoVersion.V2_0,
                type=WSchema.WappstoMetaType.NETWORK,
                id=self.uuid
            )
        )

        element = self.connection.get_network(self.uuid)
        if element:
            self.__update_self(element)
            # self.log.debug(
            #     type(self.element.meta)
            # )
            # self.log.debug(
            #     self.element.meta
            # )
            # self.log.debug(
            #     type(element.meta)
            # )
            # self.log.debug(
            #     element.meta
            # )
            if self.element != element:
                # TODO: Post diff only.
                self.log.info("Data Models Differ. Sending Local.")
                self.connection.post_network(self.element)
        else:
            self.connection.post_network(self.element)

        atexit.register(self.close)

    def clean(self) -> None:
        """
        Remove local storage, and references to the Wappsto data-model.
        """
        pass

    @property
    def name(self) -> str:
        """Returns the name of the value."""
        return self.element.name

    @property
    def uuid(self) -> UUID4:
        """Returns the name of the value."""
        return self.__uuid

    # -------------------------------------------------------------------------
    #   Helper methods
    # -------------------------------------------------------------------------

    def _setup_IoTAPI(self, configFolder, configs=None):
        # TODO: Setup the Connection.
        kwargs = self._certificate_check(configFolder)
        self.connection = IoTAPI(**kwargs)

    def _setup_RestAPI(self, configFolder, configs):
        # TODO: Setup the Connection.
        token = configs.get("token")
        login = netrc.netrc().authenticators(configs.end_point)
        if token:
            kwargs = {"token": token}
        elif login:
            kwargs = {"username": login[0], "password": login[1]}
        else:
            raise ValueError("No login was found.")
        self.connection = RestAPI(**kwargs, url=configs.end_point)

    def _device_name_gen(self, device_id):
        return f"device_{device_id}"

    def _setup_offline_storage(
        self,
        offlineStorage: Union[OfflineStorage, bool]
    ) -> None:
        # TODO: Test me!!
        if offlineStorage is False:
            return
        if offlineStorage is True:
            offline_storage: OfflineStorage = OfflineStorageFiles(
                location=self.configFolder
            )
        else:
            offline_storage: OfflineStorage = offlineStorage

        observer.subscribe(
            service.Status.SENDERROR,
            lambda _, data: offline_storage.save(data.json(exclude_none=True))
        )

        def _resend_logic(status, data):
            nonlocal offline_storage
            self.log.debug(f"Resend called with: status={status}")
            try:
                self.log.debug("Resending Offline data")
                while True:
                    data = offline_storage.load(10)
                    if not data:
                        return

                    s_data = [json.loads(d) for d in data]
                    self.log.debug(f"Sending Data: {s_data}")
                    self.connection._resend_data(
                        json.dumps(s_data)
                    )

            except Exception:
                self.log.exception("Resend Logic")

        observer.subscribe(
            connection.Status.CONNECTED,
            _resend_logic
        )

    # -------------------------------------------------------------------------
    #   Save/Load helper methods
    # -------------------------------------------------------------------------

    # def __restore_from_save_file(self, configs, kwargs):
    #     # TODO(MBK): Create the self.element & All the Children.

    #     # the kwargs weigh higher then the loaded settings.
    #     for key, value in configs.dict().items():
    #         if kwargs[key] is None:
    #             kwargs[key] = value
    #     # self.__init_devices(self.uuid, configs)
    #     pass

    def __update_self(self, element: WSchema.Network):
        # TODO(MBK): Check if new devices was added! & Check diff.
        # NOTE: If there was a diff, post local one.
        self.element = element.copy(update=self.element.dict(exclude_none=True))
        self.element.meta = element.meta.copy(update=self.element.meta)
        for nr, device in enumerate(self.element.device):
            self.cloud_id_mapping[nr] = device

    def _certificate_check(self, path) -> Dict[str, Path]:
        """
        Check if the right certificates are at the given path.
        """
        self.certi_path = {
            "ca": "ca.crt",
            "crt": "client.crt",
            "key": "client.key",
        }
        r_paths: Dict[str, Path] = {}
        for k, f in self.certi_path.items():
            r_paths[k] = path / f
            if not r_paths[k].exists():
                raise FileNotFoundError(f"'{f}' was not found in at: {path}")

        return r_paths

    # def _load_config(self) -> _ConfigFile:
    #     try:
    #         return parse_file_as(_ConfigFile, self.__config_file)
    #     except FileNotFoundError:
    #         exceptions.ConfigFileNotFoundError(
    #             "Could not find the config file."
    #         )
    #     except Exception:  # JSONDecodeError
    #         exceptions.ConfigFileError(
    #             "Could not parse config file."
    #         )

    # def _get_json(self) -> _UnitsInfo:
    #     """Generate the json-object ready for to be saved in the configfile."""
    #     unit_list = []
    #     for unit in self.children_uuid_mapping.values():
    #         unit_list.extend(unit._get_json())
    #     unit_list.append(_UnitsInfo(
    #         self_type=WSchema.WappstoObjectType.NETWORK,
    #         parent=None,
    #         children=list(self.children_uuid_mapping.keys()),
    #         children_id_mapping=self.children_id_mapping,
    #         children_name_mapping=self.children_name_mapping
    #     ))
    #     return unit_list

    # def _save_config(self):
    #     cofigdata = _ConfigFile(
    #         units=self._get_json(),
    #         configs=_Config(
    #             network_uuid=self.uuid,
    #             # network_name=self.name,
    #             port=self.port,
    #             end_point=self.end_point,
    #             # connectSync=self.connectSync,
    #             # storeQueue=self.storeQueue,
    #             # mixMaxEnforce=self.mixMaxEnforce,
    #             # stepEnforce=self.stepEnforce,
    #             # deltaHandling=self.deltaHandling,
    #             # period_handling=self.period_handling,
    #         )
    #     )
    #     with open(self.__config_file, "w") as file:
    #         file.write(cofigdata.json())

    # def __init_devices(self, uuid, configs):
    #     for device_uuid in configs.unit[uuid].children:
    #         device_settings = configs.units[device_uuid]
    #         # TODO(MBK): 'self.connection.get_device' ?
    #         theDevice = self.connection.get_device(device_uuid)
    #         temp = Device(
    #             device_id=device_settings.self_id,
    #             device_uuid=device_uuid,
    #             name=device_settings.name if device_settings.name else self._device_name_gen(device_settings.self_id),
    #             post_self=(not theDevice)
    #         )
    #         self.__add_device(device=temp, id=device_settings.self_id, name=device_settings.name)

    # def __init_values(self, uuid, parent, configs):
    #     # TODO(MBK): 'self.connection.get_value' ?    
    #     theValue = configs.units[uuid]
    #     temp = Value(
    #         name=theValue.name,
    #         value_id=theValue.self_id,
    #         value_uuid=uuid,
    #     )

    # -------------------------------------------------------------------------
    #   Status methods
    # -------------------------------------------------------------------------

    def onStatusChange(
        self,
        layer: StatusID,
        callback: Callable[[StatusID, str], None]
    ):
        """
        Configure an action when the Status have changed.

        def callback(layer: LayerEnum, newStatus: str):

        """
        observer.subscribe(
            event_name=layer,
            callback=callback
        )

    # -------------------------------------------------------------------------
    #   Network 'on-' methods
    # -------------------------------------------------------------------------

    def onChange(
        self,
        callback: Callable[[str, NetworkChangeType], None],
    ) -> None:
        """
        Configure a callback for when a change to the Network have occurred.

        # UNSURE(MBK): How should it get the data back?

        # def Networkcallback(name: str, event: NetworkChangeType, /) -> None:
        #     pass
        """
        def _cb(obj, method):
            if method == WappstoMethod.PUT:
                callback(...)

        self.connection.subscribe_network_event(
            uuid=self.uuid,
            callback=_cb
        )

    def onRequest(
        self,
        callback: Callable[[str, NetworkRequestType], None],
    ) -> None:
        """
        Configure a callback for when a request of the Network have been requested.

        # UNSURE(MBK): Name & Event, is the Same! o.0

        # def Networkcallback(name: str, event: NetworkRequestType, /) -> None:
        #     pass
        """
        def _cb(obj, method):
            if method in [WappstoMethod.DELETE, WappstoMethod.GET]:
                callback(...)

        self.connection.subscribe_network_event(
            uuid=self.uuid,
            callback=_cb
        )

    def onRefresh(
        self,
        callback: Callable[[None], None]
    ):
        """
        Configure an action when a refresh Network have been Requested.

        Normally when a refresh have been requested on a Network, ...
        ...
        # It can not! there is no '{"status":"update"}' that can be set.
        """
        def _cb(obj, method):
            if method == WappstoMethod.GET:
                callback()

        self.connection.subscribe_network_event(
            uuid=self.uuid,
            callback=_cb
        )

    def onDelete(
        self,
        callback: Callable[[None], None]
    ):
        """
        Configure an action when a Delete Network have been Requested.

        Normally when a Delete have been requested on a Network,
        it is when it is not wanted anymore, and the Network have been
        unclaimed. Which mean that all the devices & value have to be
        recreated, and/or the program have to close.
        """
        def _cb(obj, method):
            if method == WappstoMethod.DELETE:
                callback()

        self.connection.subscribe_network_event(
            uuid=self.uuid,
            callback=_cb
        )

    # -------------------------------------------------------------------------
    #   Network methods
    # -------------------------------------------------------------------------

    def change(self):
        pass

    def refresh(self):
        raise NotImplementedError("Method: 'refresh' is not Implemented.")

    def request(self):
        raise NotImplementedError("Method: 'request' is not Implemented.")

    def delete(self):
        """
        Normally it is used to unclaim a Network & delete all children.

        If a network delete itself, it will prompt a factory reset.
        This means that manufacturer and owner will be reset (or not),
        in relation of the rules set up in the certificates.
        """
        self.connection.delete_network(uuid=self.uuid)
        self._delete()

    def _delete(self):
        """Helper function for Delete, to only localy delete."""
        for c_uuid, c_obj in self.children_uuid_mapping.items():
            c_obj._delete()
        self.children_id_mapping.clear()
        self.children_name_mapping.clear()
        self.children_uuid_mapping.clear()

    # -------------------------------------------------------------------------
    #   Create methods
    # -------------------------------------------------------------------------

    def createDevice(
        self,
        name: Optional[str] = None,
        device_id: Optional[int] = None, 
        manufacturer: Optional[str] = None,
        product: Optional[str] = None,
        version: Optional[str] = None,
        serial: Optional[str] = None,
        description: Optional[str] = None,
    ) -> Device:
        """
        Create a new Wappsto Device.

        A Wappsto Device is references something that is attached to the network
        that can be controlled or have values that can be reported to Wappsto.

        This could be a button that is connected to this unit,
        or in the case of this unit is a gateway, it could be a remote unit.
        """
        kwargs = locals()
        kwargs.pop('self')

        if not device_id:
            if self.children_id_mapping:
                device_id = max(self.children_id_mapping.keys()) + 1
            else:
                device_id = 0
        elif device_id in self.children_id_mapping:
            return self.children_uuid_mapping[self.children_id_mapping[device_id]]

        kwargs['device_uuid'] = self.cloud_id_mapping.get(device_id)
        # if kwargs['device_uuid']:
        #     kwargs['name'] = self._configs.units[self.uuid].children_name_mapping.get(kwargs['device_uuid'])

        if not kwargs['name']:
            kwargs['name'] = self._device_name_gen(device_id)
        elif kwargs['name'] in self.children_name_mapping.keys():
            # The Device have already been created.
            return self.children_uuid_mapping[self.children_name_mapping[kwargs['name']]]

        device_obj = Device(parent=self, **kwargs)
        self.__add_device(device_obj, device_id, kwargs['name'])
        return device_obj

    def __add_device(self, device: Device, id: int, name: str):
        """Helper function for Create, to only localy create it."""
        self.children_uuid_mapping[device.uuid] = device
        self.children_id_mapping[id] = device.uuid
        self.children_name_mapping[name] = device.uuid

    # -------------------------------------------------------------------------
    #   Connection methods
    # -------------------------------------------------------------------------

    def connect(self):
        pass

    def disconnect(self):
        pass

    def close(self):
        """."""
        if not self.closed:
            self.connection.close()
            self.closed = True
        # Disconnect
        pass



"""This the the Simple Wappsto Python user-interface to the Wappsto devices."""

# #############################################################################
#                             Network Import Stuff
# #############################################################################

from .modules.network import Network
from .modules.network import ConnectionStatus
from .modules.network import ConnectionTypes
from .modules.network import NetworkChangeType
from .modules.network import NetworkRequestType
from .modules.network import ServiceStatus
from .modules.network import StatusID


# #############################################################################
#                             Device Import Stuff
# #############################################################################

from .modules.device import Device


# #############################################################################
#                             Value Import Stuff
# #############################################################################

from .modules.value import Value
from .modules.value import Delta
from .modules.value import Period
from .modules.value import PermissionType
from .modules.value import ValueBaseType
from .modules.template import ValueType


__version__ = "v0.5.2"



import enum
import logging

from typing import Union
from typing import Any
from typing import Callable
from typing import Optional


class Status(str, enum.Enum):
    CONNECTING = "Connecting"
    CONNECTED = "Connected"
    DISCONNECTING = "Disconnecting"
    DISCONNETCED = "Disconnected"


class Debug:
    """For debuging purposes."""
    def __init__(
        self,
        address: str,
        port: int,
        observer: Callable[[str, str], None]
    ):
        self.log = logging.getLogger(__name__)
        self.log.addHandler(logging.NullHandler())

        self.observer_name = "CONNECTION"
        self.observer = observer if observer else lambda st, nd: None
        self.observer(self.observer_name, Status.DISCONNETCED)

    def send(
        self,
        data: Union[str, bytes]
    ) -> bool:
        """
        Send the str/Bytes to the server.

        If given string, it is encoded as 'uft-8' & send.

        Returns:
            True, if the data could be send else
            False.
        """

    def receive(
        self,
        parser: Callable[[bytes], Any]
    ) -> Any:
        """
        Socket receive method.

        Method that handles receiving data from a socket. Capable of handling
        data chunks.

        Args:
            Callable: A parser, that returns the parsed data.
                      On Parsing Error, it should raise a
                      ValueError TypeError or any subClasses of those.
                      (Like 'JSONDecodeError' & 'pydantic.ValidationError' is)
        Returns:
            The Parsers output.

        """
        pass

    def connect(self) -> bool:
        """
        Connect to the server.

        Attempts a connection to the server on the provided address and port.

        Returns:
            'True' if the connection was successful else
            'False'
        """
        self.observer(self.observer_name, Status.CONNECTING)

        self.observer(self.observer_name, Status.CONNECTED)
        return True

    def reconnect(
        self,
        retry_limit: Optional[int] = None
    ) -> bool:
        """
        Attempt to reconnect.

        Close the current connection, and then try to reconnect to the server,
        until the given amount of attempts, are above the retry_limit.
        If the retry_limit are not set, it will continue end.

        Args:
            retry_limit: the amount of retries, before it stops.

        Returns:
            'True' if the connection was successful else
            'False'
        """
        self.log.info("Reconnection...")
        return True

    def disconnect(self) -> None:
        """Disconnect from the server."""

    def close(self) -> None:
        self.log.info("Closing connection...")
        self.observer(self.observer_name, Status.DISCONNECTING)
        self.observer(self.observer_name, Status.DISCONNETCED)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import json
import pathlib
import uuid

import requests

# from WappstoIoT.Modules.Template import _ConfigFile, _Config, _UnitsInfo
# from WappstoIoT.schema.iot_schema import WappstoObjectType


from rich import traceback
from rich.console import Console

console = Console()
traceback.install(show_locals=True)

debug = False

wappstoEnv = [
    "dev",
    "qa",
    "stagning",
    "prod",
]


wappstoPort = {
    "dev": 52005,
    "qa": 53005,
    "stagning": 54005,
    "prod": 443
}

wappstoUrl = {
    "dev": "https://dev.wappsto.com",
    "qa": "https://qa.wappsto.com",
    "staging": "https://stagning.wappsto.com",
    "prod": "https://wappsto.com",
}


def _log_request_error(rq):
    if debug:
        console.print("Sendt data    :")
        console.print(" - URL         : {}".format(rq.request.url))
        console.print(" - Headers     : {}".format(rq.request.headers))
        console.print(" - Request Body: {}".format(
            json.dumps(rq.request.body, indent=4, sort_keys=True))
        )

        console.print("")
        console.print("")

        console.print("Received data :")
        console.print(" - URL         : {}".format(rq.url))
        console.print(" - Headers     : {}".format(rq.headers))
        console.print(" - Status code : {}".format(rq.status_code))
        try:
            console.print(" - Request Body: {}".format(
                json.dumps(json.loads(rq.text), indent=4, sort_keys=True))
            )
        except (AttributeError, json.JSONDecodeError):

            console.print(" - Request Body: {}".format(rq.text))
    rjson = json.loads(rq.text)
    console.print(f"[bold red]{rjson['message']}")
    exit(-2)


def start_session(base_url, username, password):
    session_json = {
        "username": username,
        "password": password,
        "remember_me": False
    }

    url = f"{base_url}/services/session"

    rdata = requests.post(
        url=url,
        headers={"Content-type": "application/json"},
        data=json.dumps(session_json)
    )

    rjson = json.loads(rdata.text)

    if not rdata.ok:
        _log_request_error(rdata)

    return rjson["meta"]["id"]


def create_network(
    session,
    base_url,
    # network_uuid=None,
    product=None,
    test_mode=False,
    reset_manufacturer=False,
    manufacturer_as_owner=True
):
    # Should take use of the more general functions.
    request = {
    }
    # if network_uuid:
    #     request["network"] = {"id": uuid}
    if product:
        request["product"] = product

    if test_mode:
        request["test_mode"] = True

    if reset_manufacturer:
        request["factory_reset"] = {"reset_manufacturer": True}

    request['manufacturer_as_owner'] = manufacturer_as_owner

    url = f"{base_url}/services/2.1/creator"
    header = {
        "Content-type": "application/json",
        "X-session": str(session)
    }

    rdata = requests.post(
        url=url,
        headers=header,
        data=json.dumps(request)
    )

    rjson = json.loads(rdata.text)

    if not rdata.ok:
        _log_request_error(rdata)
    return rjson


def get_network(session, base_url, network_uuid):
    f_url = f"{base_url}/services/2.1/creator?this_network.id={network_uuid}"
    header = {
        "Content-type": "application/json",
        "X-session": str(session)
    }

    fdata = requests.get(
        url=f_url,
        headers=header,
    )

    data = json.loads(fdata.text)

    if not fdata.ok:
        _log_request_error(fdata)

    if len(data['id']) == 0:
        console.print(f"[bold red]{data['message']}")
        exit(-3)
    creator_id = data['id'][0]
    url = f"{base_url}/services/2.1/creator/{creator_id}"

    rdata = requests.get(
        url=url,
        headers=header
    )

    if not rdata.ok:
        _log_request_error(rdata)

    return json.loads(rdata.text)


def create_certificaties_files(location, creator, args):
    creator["ca"], creator["certificate"], creator["private_key"]
    with open(location / "ca.crt", "w") as file:
        file.write(creator["ca"])
    with open(location / "client.crt", "w") as file:
        file.write(creator["certificate"])
    with open(location / "client.key", "w") as file:
        file.write(creator["private_key"])


if __name__ == "__main__":

    """
    Is this even needed anymore?

    We can get the Network UUID & endpoint from the certificates.
    """

    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--env",
        type=str,
        choices=wappstoEnv,
        default="prod",
        help="Wappsto environment."
    )
    parser.add_argument(
        "--token",
        type=uuid.UUID,
        help="The Session Token. If not given, you are prompted to login."
    )
    parser.add_argument(
        "--path",
        type=pathlib.Path,
        default=".",
        help="The location to which the config files are saved."
    )
    parser.add_argument(
        "--recreate",
        type=uuid.UUID,
        help="Recreate Config file, for given network UUID. (Overwrites existent)"
    )
    parser.add_argument(
        "--debug",
        action='store_true',
        help="Make the operation more talkative",
    )

    args = parser.parse_args()

    debug = args.debug if args.debug else False

    if not args.token:
        session = start_session(
            base_url=wappstoUrl[args.env],
            username=console.input("Wappsto Username:"),
            password=console.input(prompt="Wappsto Password:", password=True),
        )
    else:
        session = args.token
    if not args.recreate:
        creator = create_network(session=session, base_url=wappstoUrl[args.env])
    else:
        creator = get_network(
            session=session,
            base_url=wappstoUrl[args.env],
            network_uuid=args.recreate,
        )

    args.path.mkdir(exist_ok=True)

    create_certificaties_files(args.path, creator, args)

    console.print(f"[bold green]New network: {creator['network']['id']}")
    console.print(f"Settings saved at: {args.path}")



import uuid
import logging

from enum import Enum

from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Union
from pydantic import UUID4

from ..service.template import ServiceClass

# from .schema.base_schema import Device as DeviceSchema
from ..schema import base_schema as WSchema
from ..schema.iot_schema import WappstoMethod
# from .schema.iot_schema import WappstoObjectType
from ..schema.base_schema import PermissionType

from .value import Value
from .template import valueSettings
from .template import ValueType
# from .Template import _UnitsInfo

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    # NOTE: To avoid ciclic import
    from .network import Network


# #############################################################################
#                                 Device Setup
# #############################################################################

class RequestType(str, Enum):
    refresh = "refresh"
    delete = "delete"


class ChangeType(str, Enum):
    value = "value"
    name = "name"
    manufacturer = "manufacturer"
    product = "product"
    version = "version"
    serial = "serial"
    description = "description"


class Device:

    schema = WSchema.Device

    def __init__(
        self,
        parent: 'Network',
        device_id: int,
        device_uuid: Optional[UUID4] = None,  # Only used on loading.
        name: Optional[str] = None,
        manufacturer: Optional[str] = None,
        product: Optional[str] = None,
        version: Optional[str] = None,
        serial: Optional[str] = None,
        description: Optional[str] = None,
        # protocol
        # communication
    ):

        self.log = logging.getLogger(__name__)
        self.log.addHandler(logging.NullHandler())

        self.parent = parent
        self.element: WSchema.Device
        self.__id: int = device_id
        self.__uuid: UUID4 = device_uuid if device_uuid else uuid.uuid4()

        self.children_uuid_mapping: Dict[UUID4, Value] = {}
        self.children_id_mapping: Dict[int, UUID4] = {}
        self.children_name_mapping: Dict[str, UUID4] = {}

        self.cloud_id_mapping: Dict[int, UUID4] = {}

        self.connection: ServiceClass = parent.connection

        self.element = self.schema(
            name=name,
            manufacturer=manufacturer,
            product=product,
            version=version,
            serial=serial,
            description=description,
            meta=WSchema.DeviceMeta(
                version=WSchema.WappstoVersion.V2_0,
                type=WSchema.WappstoMetaType.DEVICE,
                id=self.uuid
            )
        )

        element = self.connection.get_device(self.uuid)
        if element:
            self.__update_self(element)
            # self.log.debug(
            #     self.element
            # )
            # self.log.debug(
            #     element
            # )
            if self.element != element:
                # TODO: Post diff only.
                self.log.info("Data Models Differ. Sending Local.")
                self.connection.post_device(
                    network_uuid=self.parent.uuid,
                    data=self.element,
                )
        else:
            self.connection.post_device(
                network_uuid=self.parent.uuid,
                data=self.element
            )

    @property
    def name(self) -> str:
        """Returns the name of the value."""
        return self.element.name

    @property
    def uuid(self) -> UUID4:
        """Returns the name of the value."""
        return self.__uuid

    @property
    def id(self) -> int:
        return self.__id

    # -------------------------------------------------------------------------
    #   Helper methods
    # -------------------------------------------------------------------------

    # def _get_json(self) -> _UnitsInfo:
    #     """Generate the json-object ready for to be saved in the configfile."""
    #     unit_list = []
    #     for unit in self.children_uuid_mapping.values():
    #         unit_list.extend(unit._get_json())
    #     unit_list.append(_UnitsInfo(
    #         self_type=WappstoObjectType.DEVICE,
    #         self_id=self.id,
    #         parent=self.parent.uuid,
    #         children=list(self.children_uuid_mapping.keys()),
    #         children_id_mapping=self.children_id_mapping,
    #         children_name_mapping=self.children_name_mapping
    #     ))
    #     return unit_list

    def __update_self(self, element: WSchema.Device):
        # TODO(MBK): Check if new devices was added! & Check diff.
        # NOTE: If there was a diff, post local one.
        self.element = element.copy(update=self.element.dict(exclude_none=True))
        self.element.meta = element.meta.copy(update=self.element.meta)
        for nr, value in enumerate(self.element.value):
            self.cloud_id_mapping[nr] = value

    def _device_name_gen(self, name, value_id):
        return f"{name}_{value_id}"

    # -------------------------------------------------------------------------
    #   Device 'on-' methods
    # -------------------------------------------------------------------------

    def onDelete(
        self,
        callback: Callable[['Device'], None],
    ) -> None:
        """
        Configure an action when a Delete on this Device have been Requested.

        Normally when a Delete have been requested,
        it is when it is not wanted anymore.
        Which mean that all the device and it's values have to be removed,
        and the process of setting up the device, should be executed again.
        This could result in the same device are created again, or if the
        device was not found, it will just be removed.
        """
        def _cb(obj, method):
            if method in WappstoMethod.DELETE:
                callback(self)

        self.connection.subscribe_device_event(
            uuid=self.uuid,
            callback=_cb
        )

    def onRefresh(
        self,
        callback: Callable[['Device'], None],
    ) -> None:
        """
        Add trigger for when a Refresh where requested.

        # It can not! there is no '{"status":"update"}' that can be set.

        Callback:
            ValueObj: This object that have had a refresh request for.
        """
        def _cb(obj, method):
            if method in WappstoMethod.GET:
                callback(self)

        self.connection.subscribe_device_event(
            uuid=self.uuid,
            callback=_cb
        )

    def onChange(
        self,
        callback: Callable[['Device', str, Any], None],
        change_type: Optional[Union[List[ChangeType], ChangeType]] = None
    ) -> None:
        """
        Configure a callback for when a change to the Device have occurred.
        """
        pass

    def onRequest(
        self,
        callback: Callable[['Device', str, Any], None],
        request_type: Optional[RequestType] = None
    ) -> None:
        """
        Configure a callback for when a request have been make for the Device.
        """
        pass

    # -------------------------------------------------------------------------
    #   Device methods
    # -------------------------------------------------------------------------

    def change(self, change_type: ChangeType) -> None:
        """
        Update a parameter in the Device structure.

        A parameter that a device can have that can be updated could be:
         - manufacturer
         - product
         - version
         - serial
         - description
        """
        pass

    def refresh(self):
        raise NotImplementedError("Method: 'refresh' is not Implemented.")

    def request(self):
        raise NotImplementedError("Method: 'request' is not Implemented.")

    def delete(self):
        """
        Request a delete of the Device, & all it's Children.
        """
        self.connection.delete_device(uuid=self.uuid)
        self._delete()

    def _delete(self):
        for c_uuid, c_obj in self.children_uuid_mapping.items():
            c_obj._delete()
        self.children_id_mapping.clear()
        self.children_name_mapping.clear()
        self.children_uuid_mapping.clear()

    # -------------------------------------------------------------------------
    #   Other methods
    # -------------------------------------------------------------------------

    def createValue(
        self,
        name: Optional[str] = None,
        value_id: Optional[int] = None,
        value_type: ValueType = ValueType.DEFAULT,
        permission: Optional[PermissionType] = None,  # PermissionType.READWRITE,
        min: Optional[Union[int, float]] = None,
        max: Optional[Union[int, float]] = None,
        step: Optional[Union[int, float]] = None,
        encoding: Optional[str] = None,
        xsd: Optional[str] = None,
        namespace: Optional[str] = None,
        period: Optional[int] = None,  # in Sec
        delta: Optional[Union[int, float]] = None,
        description: Optional[str] = None,
        meaningful_zero: Optional[str] = None,
        mapping: Optional[bool] = None,
        ordered_mapping: Optional[bool] = None,
        si_conversion: Optional[str] = None,
        unit: Optional[str] = None,
    ) -> Value:
        """
        Create a Wappsto Value.

        A Wappsto Value is where the changing data can be found & are handled.

        If a value_type have been set, that means that the parameters like:
        name, permission, min, max, step, encoding & unit have been set
        for you, to be the right settings for the given type. But you can
        still change it, if you choose sow.
        """
        kwargs = locals()
        kwargs.pop('self')
        kwargs.pop("value_type")

        thisSetting = valueSettings[value_type]

        if not value_id:
            if self.children_id_mapping:
                value_id = self._name_gen()
            else:
                value_id = 0
        # Should we use create re-get a value?
        elif value_id in self.children_id_mapping:
            return self.children_uuid_mapping[self.children_id_mapping[value_id]]

        # the kwargs weigh higher then the default settings. 
        for key, value in thisSetting.dict().items():
            if key in kwargs and kwargs[key] is None:
                kwargs[key] = value

        kwargs['value_uuid'] = self.cloud_id_mapping.get(value_id)

        # kwargs['value_uuid'] = self.parent._configs.units[self.uuid].children_id_mapping.get(value_id)
        # if kwargs['value_uuid']:
        #     kwargs['name'] = self.parent._configs.units[self.uuid].children_name_mapping.get(kwargs['value_uuid'])

        if kwargs['name'] in self.children_name_mapping:
            # If Default value name is in use, gen new!
            kwargs['name'] = self._device_name_gen(thisSetting.name, value_id)

        value_obj = Value(
            parent=self,
            type=thisSetting.type,
            **kwargs
        )

        self.__add_value(value_obj, value_id, kwargs['name'])
        return value_obj

    def _name_gen(self) -> int:
        return max(self.children_id_mapping.keys()) + 1

    def __add_value(self, value: Value, id: int, name: str):
        """Helper function for Create, to only localy create it."""
        self.children_uuid_mapping[value.uuid] = value
        self.children_id_mapping[id] = value.uuid
        self.children_name_mapping[name] = value.uuid

    def close(self):
        pass



import uuid
import datetime
import logging

from enum import Enum

from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Union
from pydantic import UUID4

from ..service.template import ServiceClass
from .template import dict_diff
from .template import ValueBaseType
# from .template import valueSettings
from ..schema import base_schema as WSchema
from ..schema.base_schema import PermissionType
from ..schema.iot_schema import WappstoMethod

from ..utils import Timestamp

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    # NOTE: To avoid ciclic import
    from .device import Device


# #############################################################################
#                                 Value Setup
# #############################################################################


class Period(str, Enum):
    PERIODIC_REFRESH = "periodic"
    DROP_UNTIL = "drop"


class Delta(str, Enum):
    ONLY_UPDATE_IF = ""
    EXTRA_UPDATES = ""


class Value:

    class RequestType(str, Enum):
        """All the different Request types possible for a Value."""
        refresh = "refresh"
        control = "control"
        delete = "delete"

    class ChangeType(str, Enum):
        """All the different Change types possible for a Value."""
        report = "report"
        delta = "delta"
        period = "period"
        name = "name"
        description = "description"
        unit = "unit"
        min = "min"
        max = "max"
        step = "step"
        encoding = "encoding"
        meaningful_zero = "meaningful_zero"
        state = "state"  # UNSURE: ?!!?

    __value_type_2_Schema = {
        ValueBaseType.STRING: WSchema.StringValue,
        ValueBaseType.NUMBER: WSchema.NumberValue,
        ValueBaseType.BLOB: WSchema.BlobValue,
        ValueBaseType.XML: WSchema.XmlValue,
    }

    def __init__(
        self,
        parent: 'Device',
        type: ValueBaseType,
        value_id: int,
        value_uuid: Optional[UUID4] = None,  # Only used on loading.
        name: Optional[str] = None,
        permission: PermissionType = PermissionType.READWRITE,
        min: Optional[Union[int, float]] = None,
        max: Optional[Union[int, float]] = None,
        step: Optional[Union[int, float]] = None,
        encoding: Optional[str] = None,
        xsd: Optional[str] = None,
        namespace: Optional[str] = None,
        period: Optional[int] = None,  # in Sec
        delta: Optional[Union[int, float]] = None,
        description: Optional[str] = None,
        meaningful_zero: Optional[str] = None,
        mapping: Optional[bool] = None,
        ordered_mapping: Optional[bool] = None,
        si_conversion: Optional[str] = None,
        unit: Optional[str] = None,
    ):
        self.log = logging.getLogger(__name__)
        self.log.addHandler(logging.NullHandler())

        self.schema = self.__value_type_2_Schema[type]
        self.report_state: WSchema.State
        self.control_state: WSchema.State
        self.parent = parent
        self.element: Union[
            WSchema.StringValue,
            WSchema.NumberValue,
            WSchema.BlobValue,
            WSchema.XmlValue
        ] = self.schema()
        self.__id: int = value_id
        self.__uuid: UUID4 = value_uuid if value_uuid else uuid.uuid4()

        # self.children_uuid_mapping: Dict[UUID4, Value] = {}
        # self.children_id_mapping: Dict[int, UUID4] = {}
        self.children_name_mapping: Dict[str, UUID4] = {}

        self.connection: ServiceClass = parent.connection

        subValue = self.__parseValueType(
                ValueType=type,
                encoding=encoding,
                mapping=mapping,
                max_range=max,
                meaningful_zero=meaningful_zero,
                min_range=min,
                namespace=namespace,
                ordered_mapping=ordered_mapping,
                si_conversion=si_conversion,
                step=step,
                unit=unit,
                xsd=xsd,
        )

        self.element = self.schema(
            name=name,
            description=description,
            period=period,
            delta=delta,
            permission=permission,
            **subValue,
            meta=WSchema.ValueMeta(
                version=WSchema.WappstoVersion.V2_0,
                type=WSchema.WappstoMetaType.VALUE,
                id=self.uuid
            )
        )

        element = self.connection.get_value(self.uuid)

        if element:
            self.__update_self(element)
            # self.__print(element)
            if self.element != element:
                # TODO: Post diff only.
                self.log.info("Data Models Differ. Sending Local.")
                self.connection.post_value(
                    device_uuid=self.parent.uuid,
                    data=self.element
                )
            self.__update_state()
        else:
            self.connection.post_value(
                device_uuid=self.parent.uuid,
                data=self.element
            )

        self._createStates(permission)

    def __print(self, element):
        self.log.debug(
            type(self.element)
        )
        self.log.debug(
            self.element
        )
        self.log.debug(
            type(element)
        )
        self.log.debug(
            element
        )

    @property
    def data(self) -> Optional[Union[str, int, float]]:
        """
        Returns the last data value.

        The returned value will be the last Report value.
        unless there isn't one, then it will return None.
        """
        return self.report_state.data

    @property
    def target(self) -> Optional[Union[str, int, float]]:
        """
        Returns the last target value.

        The returned value will be the last Control value,
        unless there isn't one, then it will return None.
        """
        return self.control_state.data

    @property
    def name(self) -> str:
        """Returns the name of the value."""
        return self.element.name

    @property
    def uuid(self) -> UUID4:
        """Returns the uuid of the value."""
        return self.__uuid

    @property
    def id(self) -> int:
        return self.__id

    # -------------------------------------------------------------------------
    #   Helper methods
    # -------------------------------------------------------------------------

    def __parseValueType(
        self,
        ValueType,
        encoding,
        mapping,
        max_range,
        meaningful_zero,
        min_range,
        namespace,
        ordered_mapping,
        si_conversion,
        step,
        unit,
        xsd,
    ):

        if ValueType == ValueBaseType.NUMBER:
            subValue = {
                "number": WSchema.Number(
                    min=min_range,
                    max=max_range,
                    step=step,
                    mapping=mapping,
                    meaningful_zero=meaningful_zero,
                    ordered_mapping=ordered_mapping,
                    si_conversion=si_conversion,
                    unit=unit,
                )
            }
        elif ValueType == ValueBaseType.XML:
            subValue = {
                "xml": WSchema.Xml(
                    xsd=xsd,
                    namespace=namespace,
                )
            }
        elif ValueType == ValueBaseType.STRING:
            subValue = {
                "string": WSchema.String(
                    max=max_range,
                    encoding=encoding
                )
            }
        elif ValueType == ValueBaseType.BLOB:
            subValue = {
                "blob": WSchema.Blob(
                    max=max_range,
                    encoding=encoding
                )
            }

        return subValue

    # def _get_json(self) -> _UnitsInfo:
    #     """Generate the json-object ready for to be saved in the configfile."""
    #     unit_list = []
    #     for unit in self.children_uuid_mapping.values():
    #         unit_list.extend(unit._get_json())
    #     unit_list.append(_UnitsInfo(
    #         self_type=WappstoObjectType.VALUE,
    #         self_id=self.id,
    #         parent=self.parent.uuid,
    #         permission=self.element.permission,
    #         children=list(self.children_uuid_mapping.keys()),
    #         children_id_mapping=self.children_id_mapping,
    #         children_name_mapping=self.children_name_mapping
    #     ))
    #     return unit_list

    def __update_self(self, element):
        new_dict = element.copy(update=self.element.dict(exclude_none=True))
        new_dict.meta = element.meta.copy(update=new_dict.meta)
        if type(self.element) is type(element):
            self.element = new_dict
        else:
            new_dict = new_dict.dict(exclude_none=True)
            old_type = type(element)
            if old_type is WSchema.StringValue:
                new_dict.pop('string')
            elif old_type is WSchema.NumberValue:
                new_dict.pop('number')
            elif old_type is WSchema.BlobValue:
                new_dict.pop('blob')
            elif old_type is WSchema.XmlValue:
                new_dict.pop('xml')
            self.log.debug(f"CC: {new_dict}")
            self.element = self.schema(**new_dict)

        # TODO: Check for the Difference Value-types & ensure that it is right.

    def __update_state(self):
        for state_uuid in self.element.state:
            state_obj = self.connection.get_state(uuid=state_uuid)
            if state_obj:
                self.log.info(f"Found State: {state_uuid} for device: {self.uuid}")
                self.children_name_mapping[state_obj.type.name] = state_uuid
                if state_obj.type == WSchema.StateType.REPORT:
                    self.report_state = state_obj
                elif state_obj.type == WSchema.StateType.CONTROL:
                    self.control_state = state_obj

    # -------------------------------------------------------------------------
    #   Value 'on-' methods
    # -------------------------------------------------------------------------

    def onChange(
        self,
        callback: Callable[['Value', ChangeType], None],
    ) -> None:
        """
        Add a trigger on when change have been make.

        A change on the Value typically will mean that a parameter, like
        period or delta or report value have been changed,
        from the server/user side.

        Callback:
            ValueObj: The Object that have had a change to it.
            ChangeType: Name of what have change in the object.
        """
        def _cb(obj, method):
            if method in WappstoMethod.PUT:
                for key in dict_diff(self.schema.dict(), obj.dict()).keys():
                    callback(self, key)

        # UNSURE (MBK): on all state & value?
        self.connection.subscribe_value_event(
            uuid=self.uuid,
            callback=_cb
        )

    def onReport(
        self,
        callback: Callable[['Value', Union[str, int, float]], None],
    ) -> None:
        """
        Add a trigger on when Report data change have been make.

        This trigger even if the Report data was changed to the same value.

        Callback:
            Value: the Object that have had a Report for.
            Union[str, int, float]: The Value of the Report change.
        """
        raise NotImplementedError("Method: 'onReport' is not Implemented.")

        def _cb(obj, method):
            if method == WappstoMethod.PUT:
                callback(self, obj.data)

        self.connection.subscribe_state_event(
            uuid=self.children_name_mapping[WSchema.StateType.REPORT.name],
            callback=_cb
        )

    def onRequest(
        self,
        callback: Callable[['Value', RequestType, str, Any], None],
    ) -> None:
        """
        For Refresh & Control. When We are asked to be something

        # UNSURE(MBK): Name & Event, is the Same! o.0

        Callback:
            ValueObj: the Object that have had a request for.
            Event: Which type of Request have happened.
            str: Name of what to do something with.
            any: The Data.
        """
        def _cb(obj, method):
            if method in [WappstoMethod.DELETE, WappstoMethod.GET]:
                callback(...)

        # UNSURE (MBK): on all state & value?
        self.connection.subscribe_value_event(
            uuid=self.uuid,
            callback=_cb
        )

    def onControl(
        self,
        callback: Callable[['Value', Union[str, int, float]], None],
    ) -> None:
        """
        Add trigger for when a Control request have been make.

        A Control value is typical use to request a new target value,
        for the given value.

        Callback:
            ValueObj: This object that have had a request for.
            any: The Data.
        """

        def _cb(obj, method):
            if method == WappstoMethod.PUT:
                callback(self, obj.data)

        self.connection.subscribe_state_event(
            uuid=self.children_name_mapping[WSchema.StateType.CONTROL.name],
            callback=_cb
        )

    def onRefresh(
        self,
        callback: Callable[['Value'], None],
    ) -> None:
        """
        Add trigger for when a Refresh where requested.

        A Refresh is typical use to request a update of the report value,
        in case of the natural update cycle is not fast enough for the user,
        or a extra sample are wanted at that given time.

        Callback:
            ValueObj: This object that have had a refresh request for.
        """

        def _cb(obj, method):
            if method == WappstoMethod.GET:
                callback(self)

        # UNSURE (MBK): on all state & value?
        self.connection.subscribe_state_event(
            uuid=self.children_name_mapping[WSchema.StateType.REPORT.name],
            callback=_cb
        )

    def onDelete(
        self,
        callback: Callable[['Value'], None],
    ) -> None:
        """For when a 'DELETE' request have been called on this element."""
        def _cb(obj, method):
            if method == WappstoMethod.DELETE:
                callback(self)

        self.connection.subscribe_value_event(
            uuid=self.uuid,
            callback=_cb
        )

    # -------------------------------------------------------------------------
    #   Value methods
    # -------------------------------------------------------------------------

    def change(self, name: str, value: Any) -> None:
        """
        Update a parameter in the Value structure.

        A parameter that a device can have that can be updated could be:
         - Name
         - Description
         - Unit
         - min/max/step/encoding
         - period
         - delta
         - meaningful_zero
        """
        pass

    def delete(self):
        """
        Request a delete of the Device, & all it's Children.
        """

        self.connection.delete_value(uuid=self.uuid)
        # self._delete()

    # def _delete(self):
    #     # TODO: REDO!
    #     for c_uuid, c_obj in self.children_uuid_mapping.items():
    #         c_obj._delete()
    #     self.children_id_mapping.clear()
    #     self.children_name_mapping.clear()
    #     self.children_uuid_mapping.clear()

    def report(
        self,
        value: Union[int, float, str, None],
        timestamp: Optional[datetime.datetime] = None
    ) -> None:
        """
        Report the new current value to Wappsto.

        The Report value is typical a measured value from a sensor,
        whether it is a GPIO pin, a analog temperature sensor or a
        device over a I2C bus.
        """
        self.log.info(f"Sending Report for: {self.report_state.meta.id}")
        data = WSchema.State(
            data=value,
            timestamp=timestamp if timestamp else Timestamp.timestamp()
        )
        if (
            data.timestamp and self.report_state.timestamp and
            data.timestamp > self.report_state.timestamp or
            not self.report_state.timestamp
        ):
            self.report_state = self.report_state.copy(update=data.dict(exclude_none=True))

        self.connection.put_state(
            uuid=self.children_name_mapping[WSchema.StateType.REPORT.name],
            data=data
        )

    def control(
        self,
        value: Union[int, float, str, None],
        timestamp: Optional[datetime.datetime] = None
    ) -> None:
        """
        Report the a new control value to Wappsto.

        A Control value is typical only changed if a target wanted value,
        have changed, whether it is because of an on device user controller,
        or the target was outside a given range.
        """
        raise NotImplementedError("Method: 'control' is not Implemented.")

        self.log.info(f"Sending Control for: {self.control_state.meta.id}")

        data = WSchema.State(
            data=value,
            timestamp=timestamp if timestamp else Timestamp.timestamp()
        )
        if (
            data.timestamp and self.report_state.timestamp and
            data.timestamp > self.report_state.timestamp or
            not self.report_state.timestamp
        ):
            self.report_state = self.report_state.copy(update=data.dict(exclude_none=True))

        self.connection.put_state(
            uuid=self.children_name_mapping[WSchema.StateType.CONTROL.name],
            data=data
        )

    # -------------------------------------------------------------------------
    #   Other methods
    # -------------------------------------------------------------------------

    def _createStates(self, permission: PermissionType):
        if (
            permission == PermissionType.READ or
            permission == PermissionType.READWRITE
        ):
            self._CreateReport()
        if (
            permission == PermissionType.WRITE or
            permission == PermissionType.READWRITE
        ):
            self._CreateControl()

    def _CreateReport(self):
        if not self.children_name_mapping.get(WSchema.StateType.REPORT.name):
            self.children_name_mapping[WSchema.StateType.REPORT.name] = uuid.uuid4()

            self.report_state = WSchema.State(
                data=float("NaN"),
                type=WSchema.StateType.REPORT,
                meta=WSchema.BaseMeta(
                    id=self.children_name_mapping.get(WSchema.StateType.REPORT.name)
                )
            )

            self.connection.post_state(value_uuid=self.uuid, data=self.report_state)

    def _CreateControl(self):
        if not self.children_name_mapping.get(WSchema.StateType.CONTROL.name):
            self.children_name_mapping[WSchema.StateType.CONTROL.name] = uuid.uuid4()

            self.control_state = WSchema.State(
                data=float("NaN"),
                type=WSchema.StateType.CONTROL,
                meta=WSchema.BaseMeta(
                    id=self.children_name_mapping[WSchema.StateType.CONTROL.name]
                )
            )
            self.connection.post_state(value_uuid=self.uuid, data=self.control_state)

        def _cb(obj, method):
            if method == WappstoMethod.PUT:
                if obj.timestamp > self.report_state.timestamp:
                    self.log.info(f"Control Value updated: {obj.meta.id}, {obj.data}")
                    self.control_state = self.control_state.copy(update=obj.dict(exclude_none=True))

        self.connection.subscribe_state_event(
            uuid=self.children_name_mapping[WSchema.StateType.CONTROL.name],
            callback=_cb
        )

    def close(self):
        pass
